r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Autopilot
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""


from typing import Optional
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class QueryInstance(InstanceResource):
    def __init__(self, version, payload, assistant_sid: str, sid: Optional[str] = None):
        """
        Initialize the QueryInstance

        :returns: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        super().__init__(version)

        self._properties = {
            "account_sid": payload.get("account_sid"),
            "date_created": deserialize.iso8601_datetime(payload.get("date_created")),
            "date_updated": deserialize.iso8601_datetime(payload.get("date_updated")),
            "results": payload.get("results"),
            "language": payload.get("language"),
            "model_build_sid": payload.get("model_build_sid"),
            "query": payload.get("query"),
            "sample_sid": payload.get("sample_sid"),
            "assistant_sid": payload.get("assistant_sid"),
            "sid": payload.get("sid"),
            "status": payload.get("status"),
            "url": payload.get("url"),
            "source_channel": payload.get("source_channel"),
            "dialogue_sid": payload.get("dialogue_sid"),
        }

        self._solution = {
            "assistant_sid": assistant_sid,
            "sid": sid or self._properties["sid"],
        }
        self._context: Optional[QueryContext] = None

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: QueryContext for this QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryContext
        """
        if self._context is None:
            self._context = QueryContext(
                self._version,
                assistant_sid=self._solution["assistant_sid"],
                sid=self._solution["sid"],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Query resource.
        :rtype: str
        """
        return self._properties["account_sid"]

    @property
    def date_created(self):
        """
        :returns: The date and time in GMT when the resource was created specified in [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt) format.
        :rtype: datetime
        """
        return self._properties["date_created"]

    @property
    def date_updated(self):
        """
        :returns: The date and time in GMT when the resource was last updated specified in [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt) format.
        :rtype: datetime
        """
        return self._properties["date_updated"]

    @property
    def results(self):
        """
        :returns: The natural language analysis results that include the [Task](https://www.twilio.com/docs/autopilot/api/task) recognized and a list of identified [Fields](https://www.twilio.com/docs/autopilot/api/task-field).
        :rtype: dict
        """
        return self._properties["results"]

    @property
    def language(self):
        """
        :returns: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used by the Query. For example: `en-US`.
        :rtype: str
        """
        return self._properties["language"]

    @property
    def model_build_sid(self):
        """
        :returns: The SID of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) queried.
        :rtype: str
        """
        return self._properties["model_build_sid"]

    @property
    def query(self):
        """
        :returns: The end-user's natural language input.
        :rtype: str
        """
        return self._properties["query"]

    @property
    def sample_sid(self):
        """
        :returns: The SID of an optional reference to the [Sample](https://www.twilio.com/docs/autopilot/api/task-sample) created from the query.
        :rtype: str
        """
        return self._properties["sample_sid"]

    @property
    def assistant_sid(self):
        """
        :returns: The SID of the [Assistant](https://www.twilio.com/docs/autopilot/api/assistant) that is the parent of the resource.
        :rtype: str
        """
        return self._properties["assistant_sid"]

    @property
    def sid(self):
        """
        :returns: The unique string that we created to identify the Query resource.
        :rtype: str
        """
        return self._properties["sid"]

    @property
    def status(self):
        """
        :returns: The status of the Query. Can be: `pending-review`, `reviewed`, or `discarded`
        :rtype: str
        """
        return self._properties["status"]

    @property
    def url(self):
        """
        :returns: The absolute URL of the Query resource.
        :rtype: str
        """
        return self._properties["url"]

    @property
    def source_channel(self):
        """
        :returns: The communication channel from where the end-user input came.
        :rtype: str
        """
        return self._properties["source_channel"]

    @property
    def dialogue_sid(self):
        """
        :returns: The SID of the [Dialogue](https://www.twilio.com/docs/autopilot/api/dialogue).
        :rtype: str
        """
        return self._properties["dialogue_sid"]

    def delete(self):
        """
        Deletes the QueryInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._proxy.delete()

    async def delete_async(self):
        """
        Asynchronous coroutine that deletes the QueryInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return await self._proxy.delete_async()

    def fetch(self):
        """
        Fetch the QueryInstance


        :returns: The fetched QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the QueryInstance


        :returns: The fetched QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        return await self._proxy.fetch_async()

    def update(self, sample_sid=values.unset, status=values.unset):
        """
        Update the QueryInstance

        :param str sample_sid: The SID of an optional reference to the [Sample](https://www.twilio.com/docs/autopilot/api/task-sample) created from the query.
        :param str status: The new status of the resource. Can be: `pending-review`, `reviewed`, or `discarded`

        :returns: The updated QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        return self._proxy.update(
            sample_sid=sample_sid,
            status=status,
        )

    async def update_async(self, sample_sid=values.unset, status=values.unset):
        """
        Asynchronous coroutine to update the QueryInstance

        :param str sample_sid: The SID of an optional reference to the [Sample](https://www.twilio.com/docs/autopilot/api/task-sample) created from the query.
        :param str status: The new status of the resource. Can be: `pending-review`, `reviewed`, or `discarded`

        :returns: The updated QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        return await self._proxy.update_async(
            sample_sid=sample_sid,
            status=status,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Autopilot.V1.QueryInstance {}>".format(context)


class QueryContext(InstanceContext):
    def __init__(self, version: Version, assistant_sid: str, sid: str):
        """
        Initialize the QueryContext

        :param Version version: Version that contains the resource
        :param assistant_sid: The SID of the [Assistant](https://www.twilio.com/docs/autopilot/api/assistant) that is the parent of the resource to update.
        :param sid: The Twilio-provided string that uniquely identifies the Query resource to update.

        :returns: twilio.rest.autopilot.v1.assistant.query.QueryContext
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryContext
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "assistant_sid": assistant_sid,
            "sid": sid,
        }
        self._uri = "/Assistants/{assistant_sid}/Queries/{sid}".format(**self._solution)

    def delete(self):
        """
        Deletes the QueryInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._version.delete(
            method="DELETE",
            uri=self._uri,
        )

    async def delete_async(self):
        """
        Asynchronous coroutine that deletes the QueryInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return await self._version.delete_async(
            method="DELETE",
            uri=self._uri,
        )

    def fetch(self):
        """
        Fetch the QueryInstance


        :returns: The fetched QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """

        payload = self._version.fetch(
            method="GET",
            uri=self._uri,
        )

        return QueryInstance(
            self._version,
            payload,
            assistant_sid=self._solution["assistant_sid"],
            sid=self._solution["sid"],
        )

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the QueryInstance


        :returns: The fetched QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """

        payload = await self._version.fetch_async(
            method="GET",
            uri=self._uri,
        )

        return QueryInstance(
            self._version,
            payload,
            assistant_sid=self._solution["assistant_sid"],
            sid=self._solution["sid"],
        )

    def update(self, sample_sid=values.unset, status=values.unset):
        """
        Update the QueryInstance

        :param str sample_sid: The SID of an optional reference to the [Sample](https://www.twilio.com/docs/autopilot/api/task-sample) created from the query.
        :param str status: The new status of the resource. Can be: `pending-review`, `reviewed`, or `discarded`

        :returns: The updated QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        data = values.of(
            {
                "SampleSid": sample_sid,
                "Status": status,
            }
        )

        payload = self._version.update(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return QueryInstance(
            self._version,
            payload,
            assistant_sid=self._solution["assistant_sid"],
            sid=self._solution["sid"],
        )

    async def update_async(self, sample_sid=values.unset, status=values.unset):
        """
        Asynchronous coroutine to update the QueryInstance

        :param str sample_sid: The SID of an optional reference to the [Sample](https://www.twilio.com/docs/autopilot/api/task-sample) created from the query.
        :param str status: The new status of the resource. Can be: `pending-review`, `reviewed`, or `discarded`

        :returns: The updated QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        data = values.of(
            {
                "SampleSid": sample_sid,
                "Status": status,
            }
        )

        payload = await self._version.update_async(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return QueryInstance(
            self._version,
            payload,
            assistant_sid=self._solution["assistant_sid"],
            sid=self._solution["sid"],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Autopilot.V1.QueryContext {}>".format(context)


class QueryPage(Page):
    def get_instance(self, payload):
        """
        Build an instance of QueryInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        return QueryInstance(
            self._version, payload, assistant_sid=self._solution["assistant_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Autopilot.V1.QueryPage>"


class QueryList(ListResource):
    def __init__(self, version: Version, assistant_sid: str):
        """
        Initialize the QueryList

        :param Version version: Version that contains the resource
        :param assistant_sid: The SID of the [Assistant](https://www.twilio.com/docs/autopilot/api/assistant) that is the parent of the resources to read.

        :returns: twilio.rest.autopilot.v1.assistant.query.QueryList
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryList
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "assistant_sid": assistant_sid,
        }
        self._uri = "/Assistants/{assistant_sid}/Queries".format(**self._solution)

    def create(self, language, query, tasks=values.unset, model_build=values.unset):
        """
        Create the QueryInstance

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used for the new query. For example: `en-US`.
        :param str query: The end-user's natural language input. It can be up to 2048 characters long.
        :param str tasks: The list of tasks to limit the new query to. Tasks are expressed as a comma-separated list of task `unique_name` values. For example, `task-unique_name-1, task-unique_name-2`. Listing specific tasks is useful to constrain the paths that a user can take.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.

        :returns: The created QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        data = values.of(
            {
                "Language": language,
                "Query": query,
                "Tasks": tasks,
                "ModelBuild": model_build,
            }
        )

        payload = self._version.create(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return QueryInstance(
            self._version, payload, assistant_sid=self._solution["assistant_sid"]
        )

    async def create_async(
        self, language, query, tasks=values.unset, model_build=values.unset
    ):
        """
        Asynchronously create the QueryInstance

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used for the new query. For example: `en-US`.
        :param str query: The end-user's natural language input. It can be up to 2048 characters long.
        :param str tasks: The list of tasks to limit the new query to. Tasks are expressed as a comma-separated list of task `unique_name` values. For example, `task-unique_name-1, task-unique_name-2`. Listing specific tasks is useful to constrain the paths that a user can take.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.

        :returns: The created QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryInstance
        """
        data = values.of(
            {
                "Language": language,
                "Query": query,
                "Tasks": tasks,
                "ModelBuild": model_build,
            }
        )

        payload = await self._version.create_async(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return QueryInstance(
            self._version, payload, assistant_sid=self._solution["assistant_sid"]
        )

    def stream(
        self,
        language=values.unset,
        model_build=values.unset,
        status=values.unset,
        dialogue_sid=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Streams QueryInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used by the Query resources to read. For example: `en-US`.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.
        :param str status: The status of the resources to read. Can be: `pending-review`, `reviewed`, or `discarded`
        :param str dialogue_sid: The SID of the [Dialogue](https://www.twilio.com/docs/autopilot/api/dialogue).
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.autopilot.v1.assistant.query.QueryInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(
            language=language,
            model_build=model_build,
            status=status,
            dialogue_sid=dialogue_sid,
            page_size=limits["page_size"],
        )

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        language=values.unset,
        model_build=values.unset,
        status=values.unset,
        dialogue_sid=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Asynchronously streams QueryInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used by the Query resources to read. For example: `en-US`.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.
        :param str status: The status of the resources to read. Can be: `pending-review`, `reviewed`, or `discarded`
        :param str dialogue_sid: The SID of the [Dialogue](https://www.twilio.com/docs/autopilot/api/dialogue).
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.autopilot.v1.assistant.query.QueryInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            language=language,
            model_build=model_build,
            status=status,
            dialogue_sid=dialogue_sid,
            page_size=limits["page_size"],
        )

        return await self._version.stream_async(page, limits["limit"])

    def list(
        self,
        language=values.unset,
        model_build=values.unset,
        status=values.unset,
        dialogue_sid=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Lists QueryInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used by the Query resources to read. For example: `en-US`.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.
        :param str status: The status of the resources to read. Can be: `pending-review`, `reviewed`, or `discarded`
        :param str dialogue_sid: The SID of the [Dialogue](https://www.twilio.com/docs/autopilot/api/dialogue).
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.autopilot.v1.assistant.query.QueryInstance]
        """
        return list(
            self.stream(
                language=language,
                model_build=model_build,
                status=status,
                dialogue_sid=dialogue_sid,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        language=values.unset,
        model_build=values.unset,
        status=values.unset,
        dialogue_sid=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Asynchronously lists QueryInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used by the Query resources to read. For example: `en-US`.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.
        :param str status: The status of the resources to read. Can be: `pending-review`, `reviewed`, or `discarded`
        :param str dialogue_sid: The SID of the [Dialogue](https://www.twilio.com/docs/autopilot/api/dialogue).
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.autopilot.v1.assistant.query.QueryInstance]
        """
        return list(
            await self.stream_async(
                language=language,
                model_build=model_build,
                status=status,
                dialogue_sid=dialogue_sid,
                limit=limit,
                page_size=page_size,
            )
        )

    def page(
        self,
        language=values.unset,
        model_build=values.unset,
        status=values.unset,
        dialogue_sid=values.unset,
        page_token=values.unset,
        page_number=values.unset,
        page_size=values.unset,
    ):
        """
        Retrieve a single page of QueryInstance records from the API.
        Request is executed immediately

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used by the Query resources to read. For example: `en-US`.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.
        :param str status: The status of the resources to read. Can be: `pending-review`, `reviewed`, or `discarded`
        :param str dialogue_sid: The SID of the [Dialogue](https://www.twilio.com/docs/autopilot/api/dialogue).
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryPage
        """
        data = values.of(
            {
                "Language": language,
                "ModelBuild": model_build,
                "Status": status,
                "DialogueSid": dialogue_sid,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = self._version.page(method="GET", uri=self._uri, params=data)
        return QueryPage(self._version, response, self._solution)

    async def page_async(
        self,
        language=values.unset,
        model_build=values.unset,
        status=values.unset,
        dialogue_sid=values.unset,
        page_token=values.unset,
        page_number=values.unset,
        page_size=values.unset,
    ):
        """
        Asynchronously retrieve a single page of QueryInstance records from the API.
        Request is executed immediately

        :param str language: The [ISO language-country](https://docs.oracle.com/cd/E13214_01/wli/docs92/xref/xqisocodes.html) string that specifies the language used by the Query resources to read. For example: `en-US`.
        :param str model_build: The SID or unique name of the [Model Build](https://www.twilio.com/docs/autopilot/api/model-build) to be queried.
        :param str status: The status of the resources to read. Can be: `pending-review`, `reviewed`, or `discarded`
        :param str dialogue_sid: The SID of the [Dialogue](https://www.twilio.com/docs/autopilot/api/dialogue).
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryPage
        """
        data = values.of(
            {
                "Language": language,
                "ModelBuild": model_build,
                "Status": status,
                "DialogueSid": dialogue_sid,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data
        )
        return QueryPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of QueryInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryPage
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return QueryPage(self._version, response, self._solution)

    async def get_page_async(self, target_url):
        """
        Asynchronously retrieve a specific page of QueryInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of QueryInstance
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryPage
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return QueryPage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a QueryContext

        :param sid: The Twilio-provided string that uniquely identifies the Query resource to update.

        :returns: twilio.rest.autopilot.v1.assistant.query.QueryContext
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryContext
        """
        return QueryContext(
            self._version, assistant_sid=self._solution["assistant_sid"], sid=sid
        )

    def __call__(self, sid):
        """
        Constructs a QueryContext

        :param sid: The Twilio-provided string that uniquely identifies the Query resource to update.

        :returns: twilio.rest.autopilot.v1.assistant.query.QueryContext
        :rtype: twilio.rest.autopilot.v1.assistant.query.QueryContext
        """
        return QueryContext(
            self._version, assistant_sid=self._solution["assistant_sid"], sid=sid
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return "<Twilio.Autopilot.V1.QueryList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Messaging
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""


from typing import Optional
from twilio.base import deserialize, serialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class TollfreeVerificationInstance(InstanceResource):
    class OptInType(object):
        VERBAL = "VERBAL"
        WEB_FORM = "WEB_FORM"
        PAPER_FORM = "PAPER_FORM"
        VIA_TEXT = "VIA_TEXT"
        MOBILE_QR_CODE = "MOBILE_QR_CODE"

    class Status(object):
        PENDING_REVIEW = "PENDING_REVIEW"
        IN_REVIEW = "IN_REVIEW"
        TWILIO_APPROVED = "TWILIO_APPROVED"
        TWILIO_REJECTED = "TWILIO_REJECTED"

    def __init__(self, version, payload, sid: Optional[str] = None):
        """
        Initialize the TollfreeVerificationInstance

        :returns: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        super().__init__(version)

        self._properties = {
            "sid": payload.get("sid"),
            "account_sid": payload.get("account_sid"),
            "customer_profile_sid": payload.get("customer_profile_sid"),
            "trust_product_sid": payload.get("trust_product_sid"),
            "date_created": deserialize.iso8601_datetime(payload.get("date_created")),
            "date_updated": deserialize.iso8601_datetime(payload.get("date_updated")),
            "regulated_item_sid": payload.get("regulated_item_sid"),
            "business_name": payload.get("business_name"),
            "business_street_address": payload.get("business_street_address"),
            "business_street_address2": payload.get("business_street_address2"),
            "business_city": payload.get("business_city"),
            "business_state_province_region": payload.get(
                "business_state_province_region"
            ),
            "business_postal_code": payload.get("business_postal_code"),
            "business_country": payload.get("business_country"),
            "business_website": payload.get("business_website"),
            "business_contact_first_name": payload.get("business_contact_first_name"),
            "business_contact_last_name": payload.get("business_contact_last_name"),
            "business_contact_email": payload.get("business_contact_email"),
            "business_contact_phone": payload.get("business_contact_phone"),
            "notification_email": payload.get("notification_email"),
            "use_case_categories": payload.get("use_case_categories"),
            "use_case_summary": payload.get("use_case_summary"),
            "production_message_sample": payload.get("production_message_sample"),
            "opt_in_image_urls": payload.get("opt_in_image_urls"),
            "opt_in_type": payload.get("opt_in_type"),
            "message_volume": payload.get("message_volume"),
            "additional_information": payload.get("additional_information"),
            "tollfree_phone_number_sid": payload.get("tollfree_phone_number_sid"),
            "status": payload.get("status"),
            "url": payload.get("url"),
            "resource_links": payload.get("resource_links"),
            "external_reference_id": payload.get("external_reference_id"),
        }

        self._solution = {
            "sid": sid or self._properties["sid"],
        }
        self._context: Optional[TollfreeVerificationContext] = None

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: TollfreeVerificationContext for this TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationContext
        """
        if self._context is None:
            self._context = TollfreeVerificationContext(
                self._version,
                sid=self._solution["sid"],
            )
        return self._context

    @property
    def sid(self):
        """
        :returns: The unique string to identify Tollfree Verification.
        :rtype: str
        """
        return self._properties["sid"]

    @property
    def account_sid(self):
        """
        :returns: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the Tollfree Verification resource.
        :rtype: str
        """
        return self._properties["account_sid"]

    @property
    def customer_profile_sid(self):
        """
        :returns: Customer's Profile Bundle BundleSid.
        :rtype: str
        """
        return self._properties["customer_profile_sid"]

    @property
    def trust_product_sid(self):
        """
        :returns: Tollfree TrustProduct Bundle BundleSid.
        :rtype: str
        """
        return self._properties["trust_product_sid"]

    @property
    def date_created(self):
        """
        :returns: The date and time in GMT when the resource was created specified in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        :rtype: datetime
        """
        return self._properties["date_created"]

    @property
    def date_updated(self):
        """
        :returns: The date and time in GMT when the resource was last updated specified in [ISO 8601](https://en.wikipedia.org/wiki/ISO_8601) format.
        :rtype: datetime
        """
        return self._properties["date_updated"]

    @property
    def regulated_item_sid(self):
        """
        :returns: The SID of the Regulated Item.
        :rtype: str
        """
        return self._properties["regulated_item_sid"]

    @property
    def business_name(self):
        """
        :returns: The name of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_name"]

    @property
    def business_street_address(self):
        """
        :returns: The address of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_street_address"]

    @property
    def business_street_address2(self):
        """
        :returns: The address of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_street_address2"]

    @property
    def business_city(self):
        """
        :returns: The city of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_city"]

    @property
    def business_state_province_region(self):
        """
        :returns: The state/province/region of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_state_province_region"]

    @property
    def business_postal_code(self):
        """
        :returns: The postal code of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_postal_code"]

    @property
    def business_country(self):
        """
        :returns: The country of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_country"]

    @property
    def business_website(self):
        """
        :returns: The website of the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_website"]

    @property
    def business_contact_first_name(self):
        """
        :returns: The first name of the contact for the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_contact_first_name"]

    @property
    def business_contact_last_name(self):
        """
        :returns: The last name of the contact for the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_contact_last_name"]

    @property
    def business_contact_email(self):
        """
        :returns: The email address of the contact for the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_contact_email"]

    @property
    def business_contact_phone(self):
        """
        :returns: The phone number of the contact for the business or organization using the Tollfree number.
        :rtype: str
        """
        return self._properties["business_contact_phone"]

    @property
    def notification_email(self):
        """
        :returns: The email address to receive the notification about the verification result. .
        :rtype: str
        """
        return self._properties["notification_email"]

    @property
    def use_case_categories(self):
        """
        :returns: The category of the use case for the Tollfree Number. List as many are applicable..
        :rtype: list[str]
        """
        return self._properties["use_case_categories"]

    @property
    def use_case_summary(self):
        """
        :returns: Use this to further explain how messaging is used by the business or organization.
        :rtype: str
        """
        return self._properties["use_case_summary"]

    @property
    def production_message_sample(self):
        """
        :returns: An example of message content, i.e. a sample message.
        :rtype: str
        """
        return self._properties["production_message_sample"]

    @property
    def opt_in_image_urls(self):
        """
        :returns: Link to an image that shows the opt-in workflow. Multiple images allowed and must be a publicly hosted URL.
        :rtype: list[str]
        """
        return self._properties["opt_in_image_urls"]

    @property
    def opt_in_type(self):
        """
        :returns:
        :rtype: TollfreeVerificationInstance.OptInType
        """
        return self._properties["opt_in_type"]

    @property
    def message_volume(self):
        """
        :returns: Estimate monthly volume of messages from the Tollfree Number.
        :rtype: str
        """
        return self._properties["message_volume"]

    @property
    def additional_information(self):
        """
        :returns: Additional information to be provided for verification.
        :rtype: str
        """
        return self._properties["additional_information"]

    @property
    def tollfree_phone_number_sid(self):
        """
        :returns: The SID of the Phone Number associated with the Tollfree Verification.
        :rtype: str
        """
        return self._properties["tollfree_phone_number_sid"]

    @property
    def status(self):
        """
        :returns:
        :rtype: TollfreeVerificationInstance.Status
        """
        return self._properties["status"]

    @property
    def url(self):
        """
        :returns: The absolute URL of the Tollfree Verification resource.
        :rtype: str
        """
        return self._properties["url"]

    @property
    def resource_links(self):
        """
        :returns: The URLs of the documents associated with the Tollfree Verification resource.
        :rtype: dict
        """
        return self._properties["resource_links"]

    @property
    def external_reference_id(self):
        """
        :returns: An optional external reference ID supplied by customer and echoed back on status retrieval.
        :rtype: str
        """
        return self._properties["external_reference_id"]

    def fetch(self):
        """
        Fetch the TollfreeVerificationInstance


        :returns: The fetched TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the TollfreeVerificationInstance


        :returns: The fetched TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self,
        business_name=values.unset,
        business_website=values.unset,
        notification_email=values.unset,
        use_case_categories=values.unset,
        use_case_summary=values.unset,
        production_message_sample=values.unset,
        opt_in_image_urls=values.unset,
        opt_in_type=values.unset,
        message_volume=values.unset,
        business_street_address=values.unset,
        business_street_address2=values.unset,
        business_city=values.unset,
        business_state_province_region=values.unset,
        business_postal_code=values.unset,
        business_country=values.unset,
        additional_information=values.unset,
        business_contact_first_name=values.unset,
        business_contact_last_name=values.unset,
        business_contact_email=values.unset,
        business_contact_phone=values.unset,
    ):
        """
        Update the TollfreeVerificationInstance

        :param str business_name: The name of the business or organization using the Tollfree number.
        :param str business_website: The website of the business or organization using the Tollfree number.
        :param str notification_email: The email address to receive the notification about the verification result. .
        :param list[str] use_case_categories: The category of the use case for the Tollfree Number. List as many are applicable..
        :param str use_case_summary: Use this to further explain how messaging is used by the business or organization.
        :param str production_message_sample: An example of message content, i.e. a sample message.
        :param list[str] opt_in_image_urls: Link to an image that shows the opt-in workflow. Multiple images allowed and must be a publicly hosted URL.
        :param TollfreeVerificationInstance.OptInType opt_in_type:
        :param str message_volume: Estimate monthly volume of messages from the Tollfree Number.
        :param str business_street_address: The address of the business or organization using the Tollfree number.
        :param str business_street_address2: The address of the business or organization using the Tollfree number.
        :param str business_city: The city of the business or organization using the Tollfree number.
        :param str business_state_province_region: The state/province/region of the business or organization using the Tollfree number.
        :param str business_postal_code: The postal code of the business or organization using the Tollfree number.
        :param str business_country: The country of the business or organization using the Tollfree number.
        :param str additional_information: Additional information to be provided for verification.
        :param str business_contact_first_name: The first name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_last_name: The last name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_email: The email address of the contact for the business or organization using the Tollfree number.
        :param str business_contact_phone: The phone number of the contact for the business or organization using the Tollfree number.

        :returns: The updated TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        return self._proxy.update(
            business_name=business_name,
            business_website=business_website,
            notification_email=notification_email,
            use_case_categories=use_case_categories,
            use_case_summary=use_case_summary,
            production_message_sample=production_message_sample,
            opt_in_image_urls=opt_in_image_urls,
            opt_in_type=opt_in_type,
            message_volume=message_volume,
            business_street_address=business_street_address,
            business_street_address2=business_street_address2,
            business_city=business_city,
            business_state_province_region=business_state_province_region,
            business_postal_code=business_postal_code,
            business_country=business_country,
            additional_information=additional_information,
            business_contact_first_name=business_contact_first_name,
            business_contact_last_name=business_contact_last_name,
            business_contact_email=business_contact_email,
            business_contact_phone=business_contact_phone,
        )

    async def update_async(
        self,
        business_name=values.unset,
        business_website=values.unset,
        notification_email=values.unset,
        use_case_categories=values.unset,
        use_case_summary=values.unset,
        production_message_sample=values.unset,
        opt_in_image_urls=values.unset,
        opt_in_type=values.unset,
        message_volume=values.unset,
        business_street_address=values.unset,
        business_street_address2=values.unset,
        business_city=values.unset,
        business_state_province_region=values.unset,
        business_postal_code=values.unset,
        business_country=values.unset,
        additional_information=values.unset,
        business_contact_first_name=values.unset,
        business_contact_last_name=values.unset,
        business_contact_email=values.unset,
        business_contact_phone=values.unset,
    ):
        """
        Asynchronous coroutine to update the TollfreeVerificationInstance

        :param str business_name: The name of the business or organization using the Tollfree number.
        :param str business_website: The website of the business or organization using the Tollfree number.
        :param str notification_email: The email address to receive the notification about the verification result. .
        :param list[str] use_case_categories: The category of the use case for the Tollfree Number. List as many are applicable..
        :param str use_case_summary: Use this to further explain how messaging is used by the business or organization.
        :param str production_message_sample: An example of message content, i.e. a sample message.
        :param list[str] opt_in_image_urls: Link to an image that shows the opt-in workflow. Multiple images allowed and must be a publicly hosted URL.
        :param TollfreeVerificationInstance.OptInType opt_in_type:
        :param str message_volume: Estimate monthly volume of messages from the Tollfree Number.
        :param str business_street_address: The address of the business or organization using the Tollfree number.
        :param str business_street_address2: The address of the business or organization using the Tollfree number.
        :param str business_city: The city of the business or organization using the Tollfree number.
        :param str business_state_province_region: The state/province/region of the business or organization using the Tollfree number.
        :param str business_postal_code: The postal code of the business or organization using the Tollfree number.
        :param str business_country: The country of the business or organization using the Tollfree number.
        :param str additional_information: Additional information to be provided for verification.
        :param str business_contact_first_name: The first name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_last_name: The last name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_email: The email address of the contact for the business or organization using the Tollfree number.
        :param str business_contact_phone: The phone number of the contact for the business or organization using the Tollfree number.

        :returns: The updated TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        return await self._proxy.update_async(
            business_name=business_name,
            business_website=business_website,
            notification_email=notification_email,
            use_case_categories=use_case_categories,
            use_case_summary=use_case_summary,
            production_message_sample=production_message_sample,
            opt_in_image_urls=opt_in_image_urls,
            opt_in_type=opt_in_type,
            message_volume=message_volume,
            business_street_address=business_street_address,
            business_street_address2=business_street_address2,
            business_city=business_city,
            business_state_province_region=business_state_province_region,
            business_postal_code=business_postal_code,
            business_country=business_country,
            additional_information=additional_information,
            business_contact_first_name=business_contact_first_name,
            business_contact_last_name=business_contact_last_name,
            business_contact_email=business_contact_email,
            business_contact_phone=business_contact_phone,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Messaging.V1.TollfreeVerificationInstance {}>".format(context)


class TollfreeVerificationContext(InstanceContext):
    def __init__(self, version: Version, sid: str):
        """
        Initialize the TollfreeVerificationContext

        :param Version version: Version that contains the resource
        :param sid: The unique string to identify Tollfree Verification.

        :returns: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationContext
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationContext
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "sid": sid,
        }
        self._uri = "/Tollfree/Verifications/{sid}".format(**self._solution)

    def fetch(self):
        """
        Fetch the TollfreeVerificationInstance


        :returns: The fetched TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """

        payload = self._version.fetch(
            method="GET",
            uri=self._uri,
        )

        return TollfreeVerificationInstance(
            self._version,
            payload,
            sid=self._solution["sid"],
        )

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the TollfreeVerificationInstance


        :returns: The fetched TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """

        payload = await self._version.fetch_async(
            method="GET",
            uri=self._uri,
        )

        return TollfreeVerificationInstance(
            self._version,
            payload,
            sid=self._solution["sid"],
        )

    def update(
        self,
        business_name=values.unset,
        business_website=values.unset,
        notification_email=values.unset,
        use_case_categories=values.unset,
        use_case_summary=values.unset,
        production_message_sample=values.unset,
        opt_in_image_urls=values.unset,
        opt_in_type=values.unset,
        message_volume=values.unset,
        business_street_address=values.unset,
        business_street_address2=values.unset,
        business_city=values.unset,
        business_state_province_region=values.unset,
        business_postal_code=values.unset,
        business_country=values.unset,
        additional_information=values.unset,
        business_contact_first_name=values.unset,
        business_contact_last_name=values.unset,
        business_contact_email=values.unset,
        business_contact_phone=values.unset,
    ):
        """
        Update the TollfreeVerificationInstance

        :param str business_name: The name of the business or organization using the Tollfree number.
        :param str business_website: The website of the business or organization using the Tollfree number.
        :param str notification_email: The email address to receive the notification about the verification result. .
        :param list[str] use_case_categories: The category of the use case for the Tollfree Number. List as many are applicable..
        :param str use_case_summary: Use this to further explain how messaging is used by the business or organization.
        :param str production_message_sample: An example of message content, i.e. a sample message.
        :param list[str] opt_in_image_urls: Link to an image that shows the opt-in workflow. Multiple images allowed and must be a publicly hosted URL.
        :param TollfreeVerificationInstance.OptInType opt_in_type:
        :param str message_volume: Estimate monthly volume of messages from the Tollfree Number.
        :param str business_street_address: The address of the business or organization using the Tollfree number.
        :param str business_street_address2: The address of the business or organization using the Tollfree number.
        :param str business_city: The city of the business or organization using the Tollfree number.
        :param str business_state_province_region: The state/province/region of the business or organization using the Tollfree number.
        :param str business_postal_code: The postal code of the business or organization using the Tollfree number.
        :param str business_country: The country of the business or organization using the Tollfree number.
        :param str additional_information: Additional information to be provided for verification.
        :param str business_contact_first_name: The first name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_last_name: The last name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_email: The email address of the contact for the business or organization using the Tollfree number.
        :param str business_contact_phone: The phone number of the contact for the business or organization using the Tollfree number.

        :returns: The updated TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        data = values.of(
            {
                "BusinessName": business_name,
                "BusinessWebsite": business_website,
                "NotificationEmail": notification_email,
                "UseCaseCategories": serialize.map(use_case_categories, lambda e: e),
                "UseCaseSummary": use_case_summary,
                "ProductionMessageSample": production_message_sample,
                "OptInImageUrls": serialize.map(opt_in_image_urls, lambda e: e),
                "OptInType": opt_in_type,
                "MessageVolume": message_volume,
                "BusinessStreetAddress": business_street_address,
                "BusinessStreetAddress2": business_street_address2,
                "BusinessCity": business_city,
                "BusinessStateProvinceRegion": business_state_province_region,
                "BusinessPostalCode": business_postal_code,
                "BusinessCountry": business_country,
                "AdditionalInformation": additional_information,
                "BusinessContactFirstName": business_contact_first_name,
                "BusinessContactLastName": business_contact_last_name,
                "BusinessContactEmail": business_contact_email,
                "BusinessContactPhone": business_contact_phone,
            }
        )

        payload = self._version.update(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return TollfreeVerificationInstance(
            self._version, payload, sid=self._solution["sid"]
        )

    async def update_async(
        self,
        business_name=values.unset,
        business_website=values.unset,
        notification_email=values.unset,
        use_case_categories=values.unset,
        use_case_summary=values.unset,
        production_message_sample=values.unset,
        opt_in_image_urls=values.unset,
        opt_in_type=values.unset,
        message_volume=values.unset,
        business_street_address=values.unset,
        business_street_address2=values.unset,
        business_city=values.unset,
        business_state_province_region=values.unset,
        business_postal_code=values.unset,
        business_country=values.unset,
        additional_information=values.unset,
        business_contact_first_name=values.unset,
        business_contact_last_name=values.unset,
        business_contact_email=values.unset,
        business_contact_phone=values.unset,
    ):
        """
        Asynchronous coroutine to update the TollfreeVerificationInstance

        :param str business_name: The name of the business or organization using the Tollfree number.
        :param str business_website: The website of the business or organization using the Tollfree number.
        :param str notification_email: The email address to receive the notification about the verification result. .
        :param list[str] use_case_categories: The category of the use case for the Tollfree Number. List as many are applicable..
        :param str use_case_summary: Use this to further explain how messaging is used by the business or organization.
        :param str production_message_sample: An example of message content, i.e. a sample message.
        :param list[str] opt_in_image_urls: Link to an image that shows the opt-in workflow. Multiple images allowed and must be a publicly hosted URL.
        :param TollfreeVerificationInstance.OptInType opt_in_type:
        :param str message_volume: Estimate monthly volume of messages from the Tollfree Number.
        :param str business_street_address: The address of the business or organization using the Tollfree number.
        :param str business_street_address2: The address of the business or organization using the Tollfree number.
        :param str business_city: The city of the business or organization using the Tollfree number.
        :param str business_state_province_region: The state/province/region of the business or organization using the Tollfree number.
        :param str business_postal_code: The postal code of the business or organization using the Tollfree number.
        :param str business_country: The country of the business or organization using the Tollfree number.
        :param str additional_information: Additional information to be provided for verification.
        :param str business_contact_first_name: The first name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_last_name: The last name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_email: The email address of the contact for the business or organization using the Tollfree number.
        :param str business_contact_phone: The phone number of the contact for the business or organization using the Tollfree number.

        :returns: The updated TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        data = values.of(
            {
                "BusinessName": business_name,
                "BusinessWebsite": business_website,
                "NotificationEmail": notification_email,
                "UseCaseCategories": serialize.map(use_case_categories, lambda e: e),
                "UseCaseSummary": use_case_summary,
                "ProductionMessageSample": production_message_sample,
                "OptInImageUrls": serialize.map(opt_in_image_urls, lambda e: e),
                "OptInType": opt_in_type,
                "MessageVolume": message_volume,
                "BusinessStreetAddress": business_street_address,
                "BusinessStreetAddress2": business_street_address2,
                "BusinessCity": business_city,
                "BusinessStateProvinceRegion": business_state_province_region,
                "BusinessPostalCode": business_postal_code,
                "BusinessCountry": business_country,
                "AdditionalInformation": additional_information,
                "BusinessContactFirstName": business_contact_first_name,
                "BusinessContactLastName": business_contact_last_name,
                "BusinessContactEmail": business_contact_email,
                "BusinessContactPhone": business_contact_phone,
            }
        )

        payload = await self._version.update_async(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return TollfreeVerificationInstance(
            self._version, payload, sid=self._solution["sid"]
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Messaging.V1.TollfreeVerificationContext {}>".format(context)


class TollfreeVerificationPage(Page):
    def get_instance(self, payload):
        """
        Build an instance of TollfreeVerificationInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        return TollfreeVerificationInstance(self._version, payload)

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Messaging.V1.TollfreeVerificationPage>"


class TollfreeVerificationList(ListResource):
    def __init__(self, version: Version):
        """
        Initialize the TollfreeVerificationList

        :param Version version: Version that contains the resource

        :returns: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationList
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationList
        """
        super().__init__(version)

        self._uri = "/Tollfree/Verifications"

    def create(
        self,
        business_name,
        business_website,
        notification_email,
        use_case_categories,
        use_case_summary,
        production_message_sample,
        opt_in_image_urls,
        opt_in_type,
        message_volume,
        tollfree_phone_number_sid,
        customer_profile_sid=values.unset,
        business_street_address=values.unset,
        business_street_address2=values.unset,
        business_city=values.unset,
        business_state_province_region=values.unset,
        business_postal_code=values.unset,
        business_country=values.unset,
        additional_information=values.unset,
        business_contact_first_name=values.unset,
        business_contact_last_name=values.unset,
        business_contact_email=values.unset,
        business_contact_phone=values.unset,
        external_reference_id=values.unset,
    ):
        """
        Create the TollfreeVerificationInstance

        :param str business_name: The name of the business or organization using the Tollfree number.
        :param str business_website: The website of the business or organization using the Tollfree number.
        :param str notification_email: The email address to receive the notification about the verification result. .
        :param list[str] use_case_categories: The category of the use case for the Tollfree Number. List as many are applicable..
        :param str use_case_summary: Use this to further explain how messaging is used by the business or organization.
        :param str production_message_sample: An example of message content, i.e. a sample message.
        :param list[str] opt_in_image_urls: Link to an image that shows the opt-in workflow. Multiple images allowed and must be a publicly hosted URL.
        :param TollfreeVerificationInstance.OptInType opt_in_type:
        :param str message_volume: Estimate monthly volume of messages from the Tollfree Number.
        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param str customer_profile_sid: Customer's Profile Bundle BundleSid.
        :param str business_street_address: The address of the business or organization using the Tollfree number.
        :param str business_street_address2: The address of the business or organization using the Tollfree number.
        :param str business_city: The city of the business or organization using the Tollfree number.
        :param str business_state_province_region: The state/province/region of the business or organization using the Tollfree number.
        :param str business_postal_code: The postal code of the business or organization using the Tollfree number.
        :param str business_country: The country of the business or organization using the Tollfree number.
        :param str additional_information: Additional information to be provided for verification.
        :param str business_contact_first_name: The first name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_last_name: The last name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_email: The email address of the contact for the business or organization using the Tollfree number.
        :param str business_contact_phone: The phone number of the contact for the business or organization using the Tollfree number.
        :param str external_reference_id: An optional external reference ID supplied by customer and echoed back on status retrieval.

        :returns: The created TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        data = values.of(
            {
                "BusinessName": business_name,
                "BusinessWebsite": business_website,
                "NotificationEmail": notification_email,
                "UseCaseCategories": serialize.map(use_case_categories, lambda e: e),
                "UseCaseSummary": use_case_summary,
                "ProductionMessageSample": production_message_sample,
                "OptInImageUrls": serialize.map(opt_in_image_urls, lambda e: e),
                "OptInType": opt_in_type,
                "MessageVolume": message_volume,
                "TollfreePhoneNumberSid": tollfree_phone_number_sid,
                "CustomerProfileSid": customer_profile_sid,
                "BusinessStreetAddress": business_street_address,
                "BusinessStreetAddress2": business_street_address2,
                "BusinessCity": business_city,
                "BusinessStateProvinceRegion": business_state_province_region,
                "BusinessPostalCode": business_postal_code,
                "BusinessCountry": business_country,
                "AdditionalInformation": additional_information,
                "BusinessContactFirstName": business_contact_first_name,
                "BusinessContactLastName": business_contact_last_name,
                "BusinessContactEmail": business_contact_email,
                "BusinessContactPhone": business_contact_phone,
                "ExternalReferenceId": external_reference_id,
            }
        )

        payload = self._version.create(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return TollfreeVerificationInstance(self._version, payload)

    async def create_async(
        self,
        business_name,
        business_website,
        notification_email,
        use_case_categories,
        use_case_summary,
        production_message_sample,
        opt_in_image_urls,
        opt_in_type,
        message_volume,
        tollfree_phone_number_sid,
        customer_profile_sid=values.unset,
        business_street_address=values.unset,
        business_street_address2=values.unset,
        business_city=values.unset,
        business_state_province_region=values.unset,
        business_postal_code=values.unset,
        business_country=values.unset,
        additional_information=values.unset,
        business_contact_first_name=values.unset,
        business_contact_last_name=values.unset,
        business_contact_email=values.unset,
        business_contact_phone=values.unset,
        external_reference_id=values.unset,
    ):
        """
        Asynchronously create the TollfreeVerificationInstance

        :param str business_name: The name of the business or organization using the Tollfree number.
        :param str business_website: The website of the business or organization using the Tollfree number.
        :param str notification_email: The email address to receive the notification about the verification result. .
        :param list[str] use_case_categories: The category of the use case for the Tollfree Number. List as many are applicable..
        :param str use_case_summary: Use this to further explain how messaging is used by the business or organization.
        :param str production_message_sample: An example of message content, i.e. a sample message.
        :param list[str] opt_in_image_urls: Link to an image that shows the opt-in workflow. Multiple images allowed and must be a publicly hosted URL.
        :param TollfreeVerificationInstance.OptInType opt_in_type:
        :param str message_volume: Estimate monthly volume of messages from the Tollfree Number.
        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param str customer_profile_sid: Customer's Profile Bundle BundleSid.
        :param str business_street_address: The address of the business or organization using the Tollfree number.
        :param str business_street_address2: The address of the business or organization using the Tollfree number.
        :param str business_city: The city of the business or organization using the Tollfree number.
        :param str business_state_province_region: The state/province/region of the business or organization using the Tollfree number.
        :param str business_postal_code: The postal code of the business or organization using the Tollfree number.
        :param str business_country: The country of the business or organization using the Tollfree number.
        :param str additional_information: Additional information to be provided for verification.
        :param str business_contact_first_name: The first name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_last_name: The last name of the contact for the business or organization using the Tollfree number.
        :param str business_contact_email: The email address of the contact for the business or organization using the Tollfree number.
        :param str business_contact_phone: The phone number of the contact for the business or organization using the Tollfree number.
        :param str external_reference_id: An optional external reference ID supplied by customer and echoed back on status retrieval.

        :returns: The created TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance
        """
        data = values.of(
            {
                "BusinessName": business_name,
                "BusinessWebsite": business_website,
                "NotificationEmail": notification_email,
                "UseCaseCategories": serialize.map(use_case_categories, lambda e: e),
                "UseCaseSummary": use_case_summary,
                "ProductionMessageSample": production_message_sample,
                "OptInImageUrls": serialize.map(opt_in_image_urls, lambda e: e),
                "OptInType": opt_in_type,
                "MessageVolume": message_volume,
                "TollfreePhoneNumberSid": tollfree_phone_number_sid,
                "CustomerProfileSid": customer_profile_sid,
                "BusinessStreetAddress": business_street_address,
                "BusinessStreetAddress2": business_street_address2,
                "BusinessCity": business_city,
                "BusinessStateProvinceRegion": business_state_province_region,
                "BusinessPostalCode": business_postal_code,
                "BusinessCountry": business_country,
                "AdditionalInformation": additional_information,
                "BusinessContactFirstName": business_contact_first_name,
                "BusinessContactLastName": business_contact_last_name,
                "BusinessContactEmail": business_contact_email,
                "BusinessContactPhone": business_contact_phone,
                "ExternalReferenceId": external_reference_id,
            }
        )

        payload = await self._version.create_async(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return TollfreeVerificationInstance(self._version, payload)

    def stream(
        self,
        tollfree_phone_number_sid=values.unset,
        status=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Streams TollfreeVerificationInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param TollfreeVerificationInstance.Status status: The compliance status of the Tollfree Verification record.
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(
            tollfree_phone_number_sid=tollfree_phone_number_sid,
            status=status,
            page_size=limits["page_size"],
        )

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        tollfree_phone_number_sid=values.unset,
        status=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Asynchronously streams TollfreeVerificationInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param TollfreeVerificationInstance.Status status: The compliance status of the Tollfree Verification record.
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            tollfree_phone_number_sid=tollfree_phone_number_sid,
            status=status,
            page_size=limits["page_size"],
        )

        return await self._version.stream_async(page, limits["limit"])

    def list(
        self,
        tollfree_phone_number_sid=values.unset,
        status=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Lists TollfreeVerificationInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param TollfreeVerificationInstance.Status status: The compliance status of the Tollfree Verification record.
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance]
        """
        return list(
            self.stream(
                tollfree_phone_number_sid=tollfree_phone_number_sid,
                status=status,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        tollfree_phone_number_sid=values.unset,
        status=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Asynchronously lists TollfreeVerificationInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param TollfreeVerificationInstance.Status status: The compliance status of the Tollfree Verification record.
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationInstance]
        """
        return list(
            await self.stream_async(
                tollfree_phone_number_sid=tollfree_phone_number_sid,
                status=status,
                limit=limit,
                page_size=page_size,
            )
        )

    def page(
        self,
        tollfree_phone_number_sid=values.unset,
        status=values.unset,
        page_token=values.unset,
        page_number=values.unset,
        page_size=values.unset,
    ):
        """
        Retrieve a single page of TollfreeVerificationInstance records from the API.
        Request is executed immediately

        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param TollfreeVerificationInstance.Status status: The compliance status of the Tollfree Verification record.
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationPage
        """
        data = values.of(
            {
                "TollfreePhoneNumberSid": tollfree_phone_number_sid,
                "Status": status,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = self._version.page(method="GET", uri=self._uri, params=data)
        return TollfreeVerificationPage(self._version, response)

    async def page_async(
        self,
        tollfree_phone_number_sid=values.unset,
        status=values.unset,
        page_token=values.unset,
        page_number=values.unset,
        page_size=values.unset,
    ):
        """
        Asynchronously retrieve a single page of TollfreeVerificationInstance records from the API.
        Request is executed immediately

        :param str tollfree_phone_number_sid: The SID of the Phone Number associated with the Tollfree Verification.
        :param TollfreeVerificationInstance.Status status: The compliance status of the Tollfree Verification record.
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationPage
        """
        data = values.of(
            {
                "TollfreePhoneNumberSid": tollfree_phone_number_sid,
                "Status": status,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data
        )
        return TollfreeVerificationPage(self._version, response)

    def get_page(self, target_url):
        """
        Retrieve a specific page of TollfreeVerificationInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationPage
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return TollfreeVerificationPage(self._version, response)

    async def get_page_async(self, target_url):
        """
        Asynchronously retrieve a specific page of TollfreeVerificationInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of TollfreeVerificationInstance
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationPage
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return TollfreeVerificationPage(self._version, response)

    def get(self, sid):
        """
        Constructs a TollfreeVerificationContext

        :param sid: The unique string to identify Tollfree Verification.

        :returns: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationContext
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationContext
        """
        return TollfreeVerificationContext(self._version, sid=sid)

    def __call__(self, sid):
        """
        Constructs a TollfreeVerificationContext

        :param sid: The unique string to identify Tollfree Verification.

        :returns: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationContext
        :rtype: twilio.rest.messaging.v1.tollfree_verification.TollfreeVerificationContext
        """
        return TollfreeVerificationContext(self._version, sid=sid)

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return "<Twilio.Messaging.V1.TollfreeVerificationList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Messaging
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""


from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version


class UsecaseInstance(InstanceResource):
    def __init__(self, version, payload):
        """
        Initialize the UsecaseInstance

        :returns: twilio.rest.messaging.v1.usecase.UsecaseInstance
        :rtype: twilio.rest.messaging.v1.usecase.UsecaseInstance
        """
        super().__init__(version)

        self._properties = {
            "usecases": payload.get("usecases"),
        }

        self._solution = {}

    @property
    def usecases(self):
        """
        :returns: Human readable use case details (usecase, description and purpose) of Messaging Service Use Cases.
        :rtype: list[object]
        """
        return self._properties["usecases"]

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Messaging.V1.UsecaseInstance {}>".format(context)


class UsecaseList(ListResource):
    def __init__(self, version: Version):
        """
        Initialize the UsecaseList

        :param Version version: Version that contains the resource

        :returns: twilio.rest.messaging.v1.usecase.UsecaseList
        :rtype: twilio.rest.messaging.v1.usecase.UsecaseList
        """
        super().__init__(version)

        self._uri = "/Services/Usecases"

    def fetch(self):
        """
        Asynchronously fetch the UsecaseInstance

        :returns: The fetched UsecaseInstance
        :rtype: twilio.rest.messaging.v1.usecase.UsecaseInstance
        """
        payload = self._version.fetch(method="GET", uri=self._uri)

        return UsecaseInstance(self._version, payload)

    async def fetch_async(self):
        """
        Asynchronously fetch the UsecaseInstance

        :returns: The fetched UsecaseInstance
        :rtype: twilio.rest.messaging.v1.usecase.UsecaseInstance
        """
        payload = await self._version.fetch_async(method="GET", uri=self._uri)

        return UsecaseInstance(self._version, payload)

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return "<Twilio.Messaging.V1.UsecaseList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Api
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""


from typing import Optional
from twilio.base import serialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class ConnectAppInstance(InstanceResource):
    class Permission(object):
        GET_ALL = "get-all"
        POST_ALL = "post-all"

    def __init__(self, version, payload, account_sid: str, sid: Optional[str] = None):
        """
        Initialize the ConnectAppInstance

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        super().__init__(version)

        self._properties = {
            "account_sid": payload.get("account_sid"),
            "authorize_redirect_url": payload.get("authorize_redirect_url"),
            "company_name": payload.get("company_name"),
            "deauthorize_callback_method": payload.get("deauthorize_callback_method"),
            "deauthorize_callback_url": payload.get("deauthorize_callback_url"),
            "description": payload.get("description"),
            "friendly_name": payload.get("friendly_name"),
            "homepage_url": payload.get("homepage_url"),
            "permissions": payload.get("permissions"),
            "sid": payload.get("sid"),
            "uri": payload.get("uri"),
        }

        self._solution = {
            "account_sid": account_sid,
            "sid": sid or self._properties["sid"],
        }
        self._context: Optional[ConnectAppContext] = None

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: ConnectAppContext for this ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        if self._context is None:
            self._context = ConnectAppContext(
                self._version,
                account_sid=self._solution["account_sid"],
                sid=self._solution["sid"],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the ConnectApp resource.
        :rtype: str
        """
        return self._properties["account_sid"]

    @property
    def authorize_redirect_url(self):
        """
        :returns: The URL we redirect the user to after we authenticate the user and obtain authorization to access the Connect App.
        :rtype: str
        """
        return self._properties["authorize_redirect_url"]

    @property
    def company_name(self):
        """
        :returns: The company name set for the Connect App.
        :rtype: str
        """
        return self._properties["company_name"]

    @property
    def deauthorize_callback_method(self):
        """
        :returns: The HTTP method we use to call `deauthorize_callback_url`.
        :rtype: str
        """
        return self._properties["deauthorize_callback_method"]

    @property
    def deauthorize_callback_url(self):
        """
        :returns: The URL we call using the `deauthorize_callback_method` to de-authorize the Connect App.
        :rtype: str
        """
        return self._properties["deauthorize_callback_url"]

    @property
    def description(self):
        """
        :returns: The description of the Connect App.
        :rtype: str
        """
        return self._properties["description"]

    @property
    def friendly_name(self):
        """
        :returns: The string that you assigned to describe the resource.
        :rtype: str
        """
        return self._properties["friendly_name"]

    @property
    def homepage_url(self):
        """
        :returns: The public URL where users can obtain more information about this Connect App.
        :rtype: str
        """
        return self._properties["homepage_url"]

    @property
    def permissions(self):
        """
        :returns: The set of permissions that your ConnectApp requests.
        :rtype: list[ConnectAppInstance.Permission]
        """
        return self._properties["permissions"]

    @property
    def sid(self):
        """
        :returns: The unique string that that we created to identify the ConnectApp resource.
        :rtype: str
        """
        return self._properties["sid"]

    @property
    def uri(self):
        """
        :returns: The URI of the resource, relative to `https://api.twilio.com`.
        :rtype: str
        """
        return self._properties["uri"]

    def delete(self):
        """
        Deletes the ConnectAppInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._proxy.delete()

    async def delete_async(self):
        """
        Asynchronous coroutine that deletes the ConnectAppInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return await self._proxy.delete_async()

    def fetch(self):
        """
        Fetch the ConnectAppInstance


        :returns: The fetched ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the ConnectAppInstance


        :returns: The fetched ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self,
        authorize_redirect_url=values.unset,
        company_name=values.unset,
        deauthorize_callback_method=values.unset,
        deauthorize_callback_url=values.unset,
        description=values.unset,
        friendly_name=values.unset,
        homepage_url=values.unset,
        permissions=values.unset,
    ):
        """
        Update the ConnectAppInstance

        :param str authorize_redirect_url: The URL to redirect the user to after we authenticate the user and obtain authorization to access the Connect App.
        :param str company_name: The company name to set for the Connect App.
        :param str deauthorize_callback_method: The HTTP method to use when calling `deauthorize_callback_url`.
        :param str deauthorize_callback_url: The URL to call using the `deauthorize_callback_method` to de-authorize the Connect App.
        :param str description: A description of the Connect App.
        :param str friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.
        :param str homepage_url: A public URL where users can obtain more information about this Connect App.
        :param list[ConnectAppInstance.Permission] permissions: A comma-separated list of the permissions you will request from the users of this ConnectApp.  Can include: `get-all` and `post-all`.

        :returns: The updated ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return self._proxy.update(
            authorize_redirect_url=authorize_redirect_url,
            company_name=company_name,
            deauthorize_callback_method=deauthorize_callback_method,
            deauthorize_callback_url=deauthorize_callback_url,
            description=description,
            friendly_name=friendly_name,
            homepage_url=homepage_url,
            permissions=permissions,
        )

    async def update_async(
        self,
        authorize_redirect_url=values.unset,
        company_name=values.unset,
        deauthorize_callback_method=values.unset,
        deauthorize_callback_url=values.unset,
        description=values.unset,
        friendly_name=values.unset,
        homepage_url=values.unset,
        permissions=values.unset,
    ):
        """
        Asynchronous coroutine to update the ConnectAppInstance

        :param str authorize_redirect_url: The URL to redirect the user to after we authenticate the user and obtain authorization to access the Connect App.
        :param str company_name: The company name to set for the Connect App.
        :param str deauthorize_callback_method: The HTTP method to use when calling `deauthorize_callback_url`.
        :param str deauthorize_callback_url: The URL to call using the `deauthorize_callback_method` to de-authorize the Connect App.
        :param str description: A description of the Connect App.
        :param str friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.
        :param str homepage_url: A public URL where users can obtain more information about this Connect App.
        :param list[ConnectAppInstance.Permission] permissions: A comma-separated list of the permissions you will request from the users of this ConnectApp.  Can include: `get-all` and `post-all`.

        :returns: The updated ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return await self._proxy.update_async(
            authorize_redirect_url=authorize_redirect_url,
            company_name=company_name,
            deauthorize_callback_method=deauthorize_callback_method,
            deauthorize_callback_url=deauthorize_callback_url,
            description=description,
            friendly_name=friendly_name,
            homepage_url=homepage_url,
            permissions=permissions,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.ConnectAppInstance {}>".format(context)


class ConnectAppContext(InstanceContext):
    def __init__(self, version: Version, account_sid: str, sid: str):
        """
        Initialize the ConnectAppContext

        :param Version version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the ConnectApp resources to update.
        :param sid: The Twilio-provided string that uniquely identifies the ConnectApp resource to update.

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
            "sid": sid,
        }
        self._uri = "/Accounts/{account_sid}/ConnectApps/{sid}.json".format(
            **self._solution
        )

    def delete(self):
        """
        Deletes the ConnectAppInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._version.delete(
            method="DELETE",
            uri=self._uri,
        )

    async def delete_async(self):
        """
        Asynchronous coroutine that deletes the ConnectAppInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return await self._version.delete_async(
            method="DELETE",
            uri=self._uri,
        )

    def fetch(self):
        """
        Fetch the ConnectAppInstance


        :returns: The fetched ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """

        payload = self._version.fetch(
            method="GET",
            uri=self._uri,
        )

        return ConnectAppInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the ConnectAppInstance


        :returns: The fetched ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """

        payload = await self._version.fetch_async(
            method="GET",
            uri=self._uri,
        )

        return ConnectAppInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    def update(
        self,
        authorize_redirect_url=values.unset,
        company_name=values.unset,
        deauthorize_callback_method=values.unset,
        deauthorize_callback_url=values.unset,
        description=values.unset,
        friendly_name=values.unset,
        homepage_url=values.unset,
        permissions=values.unset,
    ):
        """
        Update the ConnectAppInstance

        :param str authorize_redirect_url: The URL to redirect the user to after we authenticate the user and obtain authorization to access the Connect App.
        :param str company_name: The company name to set for the Connect App.
        :param str deauthorize_callback_method: The HTTP method to use when calling `deauthorize_callback_url`.
        :param str deauthorize_callback_url: The URL to call using the `deauthorize_callback_method` to de-authorize the Connect App.
        :param str description: A description of the Connect App.
        :param str friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.
        :param str homepage_url: A public URL where users can obtain more information about this Connect App.
        :param list[ConnectAppInstance.Permission] permissions: A comma-separated list of the permissions you will request from the users of this ConnectApp.  Can include: `get-all` and `post-all`.

        :returns: The updated ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        data = values.of(
            {
                "AuthorizeRedirectUrl": authorize_redirect_url,
                "CompanyName": company_name,
                "DeauthorizeCallbackMethod": deauthorize_callback_method,
                "DeauthorizeCallbackUrl": deauthorize_callback_url,
                "Description": description,
                "FriendlyName": friendly_name,
                "HomepageUrl": homepage_url,
                "Permissions": serialize.map(permissions, lambda e: e),
            }
        )

        payload = self._version.update(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return ConnectAppInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    async def update_async(
        self,
        authorize_redirect_url=values.unset,
        company_name=values.unset,
        deauthorize_callback_method=values.unset,
        deauthorize_callback_url=values.unset,
        description=values.unset,
        friendly_name=values.unset,
        homepage_url=values.unset,
        permissions=values.unset,
    ):
        """
        Asynchronous coroutine to update the ConnectAppInstance

        :param str authorize_redirect_url: The URL to redirect the user to after we authenticate the user and obtain authorization to access the Connect App.
        :param str company_name: The company name to set for the Connect App.
        :param str deauthorize_callback_method: The HTTP method to use when calling `deauthorize_callback_url`.
        :param str deauthorize_callback_url: The URL to call using the `deauthorize_callback_method` to de-authorize the Connect App.
        :param str description: A description of the Connect App.
        :param str friendly_name: A descriptive string that you create to describe the resource. It can be up to 64 characters long.
        :param str homepage_url: A public URL where users can obtain more information about this Connect App.
        :param list[ConnectAppInstance.Permission] permissions: A comma-separated list of the permissions you will request from the users of this ConnectApp.  Can include: `get-all` and `post-all`.

        :returns: The updated ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        data = values.of(
            {
                "AuthorizeRedirectUrl": authorize_redirect_url,
                "CompanyName": company_name,
                "DeauthorizeCallbackMethod": deauthorize_callback_method,
                "DeauthorizeCallbackUrl": deauthorize_callback_url,
                "Description": description,
                "FriendlyName": friendly_name,
                "HomepageUrl": homepage_url,
                "Permissions": serialize.map(permissions, lambda e: e),
            }
        )

        payload = await self._version.update_async(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return ConnectAppInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.ConnectAppContext {}>".format(context)


class ConnectAppPage(Page):
    def get_instance(self, payload):
        """
        Build an instance of ConnectAppInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return ConnectAppInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.ConnectAppPage>"


class ConnectAppList(ListResource):
    def __init__(self, version: Version, account_sid: str):
        """
        Initialize the ConnectAppList

        :param Version version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that created the ConnectApp resources to read.

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppList
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppList
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
        }
        self._uri = "/Accounts/{account_sid}/ConnectApps.json".format(**self._solution)

    def stream(self, limit=None, page_size=None):
        """
        Streams ConnectAppInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.connect_app.ConnectAppInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(self, limit=None, page_size=None):
        """
        Asynchronously streams ConnectAppInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.connect_app.ConnectAppInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(page_size=limits["page_size"])

        return await self._version.stream_async(page, limits["limit"])

    def list(self, limit=None, page_size=None):
        """
        Lists ConnectAppInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.connect_app.ConnectAppInstance]
        """
        return list(
            self.stream(
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(self, limit=None, page_size=None):
        """
        Asynchronously lists ConnectAppInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.connect_app.ConnectAppInstance]
        """
        return list(
            await self.stream_async(
                limit=limit,
                page_size=page_size,
            )
        )

    def page(
        self, page_token=values.unset, page_number=values.unset, page_size=values.unset
    ):
        """
        Retrieve a single page of ConnectAppInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = self._version.page(method="GET", uri=self._uri, params=data)
        return ConnectAppPage(self._version, response, self._solution)

    async def page_async(
        self, page_token=values.unset, page_number=values.unset, page_size=values.unset
    ):
        """
        Asynchronously retrieve a single page of ConnectAppInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data
        )
        return ConnectAppPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of ConnectAppInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return ConnectAppPage(self._version, response, self._solution)

    async def get_page_async(self, target_url):
        """
        Asynchronously retrieve a specific page of ConnectAppInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return ConnectAppPage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a ConnectAppContext

        :param sid: The Twilio-provided string that uniquely identifies the ConnectApp resource to update.

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        return ConnectAppContext(
            self._version, account_sid=self._solution["account_sid"], sid=sid
        )

    def __call__(self, sid):
        """
        Constructs a ConnectAppContext

        :param sid: The Twilio-provided string that uniquely identifies the ConnectApp resource to update.

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        return ConnectAppContext(
            self._version, account_sid=self._solution["account_sid"], sid=sid
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return "<Twilio.Api.V2010.ConnectAppList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Api
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""


from typing import Optional
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page
from twilio.rest.api.v2010.account.address.dependent_phone_number import (
    DependentPhoneNumberList,
)


class AddressInstance(InstanceResource):
    def __init__(self, version, payload, account_sid: str, sid: Optional[str] = None):
        """
        Initialize the AddressInstance

        :returns: twilio.rest.api.v2010.account.address.AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        super().__init__(version)

        self._properties = {
            "account_sid": payload.get("account_sid"),
            "city": payload.get("city"),
            "customer_name": payload.get("customer_name"),
            "date_created": deserialize.rfc2822_datetime(payload.get("date_created")),
            "date_updated": deserialize.rfc2822_datetime(payload.get("date_updated")),
            "friendly_name": payload.get("friendly_name"),
            "iso_country": payload.get("iso_country"),
            "postal_code": payload.get("postal_code"),
            "region": payload.get("region"),
            "sid": payload.get("sid"),
            "street": payload.get("street"),
            "uri": payload.get("uri"),
            "emergency_enabled": payload.get("emergency_enabled"),
            "validated": payload.get("validated"),
            "verified": payload.get("verified"),
            "street_secondary": payload.get("street_secondary"),
        }

        self._solution = {
            "account_sid": account_sid,
            "sid": sid or self._properties["sid"],
        }
        self._context: Optional[AddressContext] = None

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: AddressContext for this AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressContext
        """
        if self._context is None:
            self._context = AddressContext(
                self._version,
                account_sid=self._solution["account_sid"],
                sid=self._solution["sid"],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that is responsible for the Address resource.
        :rtype: str
        """
        return self._properties["account_sid"]

    @property
    def city(self):
        """
        :returns: The city in which the address is located.
        :rtype: str
        """
        return self._properties["city"]

    @property
    def customer_name(self):
        """
        :returns: The name associated with the address.This property has a maximum length of 16 4-byte characters, or 21 3-byte characters.
        :rtype: str
        """
        return self._properties["customer_name"]

    @property
    def date_created(self):
        """
        :returns: The date and time in GMT that the resource was created specified in [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt) format.
        :rtype: datetime
        """
        return self._properties["date_created"]

    @property
    def date_updated(self):
        """
        :returns: The date and time in GMT that the resource was last updated specified in [RFC 2822](https://www.ietf.org/rfc/rfc2822.txt) format.
        :rtype: datetime
        """
        return self._properties["date_updated"]

    @property
    def friendly_name(self):
        """
        :returns: The string that you assigned to describe the resource.
        :rtype: str
        """
        return self._properties["friendly_name"]

    @property
    def iso_country(self):
        """
        :returns: The ISO country code of the address.
        :rtype: str
        """
        return self._properties["iso_country"]

    @property
    def postal_code(self):
        """
        :returns: The postal code of the address.
        :rtype: str
        """
        return self._properties["postal_code"]

    @property
    def region(self):
        """
        :returns: The state or region of the address.
        :rtype: str
        """
        return self._properties["region"]

    @property
    def sid(self):
        """
        :returns: The unique string that that we created to identify the Address resource.
        :rtype: str
        """
        return self._properties["sid"]

    @property
    def street(self):
        """
        :returns: The number and street address of the address.
        :rtype: str
        """
        return self._properties["street"]

    @property
    def uri(self):
        """
        :returns: The URI of the resource, relative to `https://api.twilio.com`.
        :rtype: str
        """
        return self._properties["uri"]

    @property
    def emergency_enabled(self):
        """
        :returns: Whether emergency calling has been enabled on this number.
        :rtype: bool
        """
        return self._properties["emergency_enabled"]

    @property
    def validated(self):
        """
        :returns: Whether the address has been validated to comply with local regulation. In countries that require valid addresses, an invalid address will not be accepted. `true` indicates the Address has been validated. `false` indicate the country doesn't require validation or the Address is not valid.
        :rtype: bool
        """
        return self._properties["validated"]

    @property
    def verified(self):
        """
        :returns: Whether the address has been verified to comply with regulation. In countries that require valid addresses, an invalid address will not be accepted. `true` indicates the Address has been verified. `false` indicate the country doesn't require verified or the Address is not valid.
        :rtype: bool
        """
        return self._properties["verified"]

    @property
    def street_secondary(self):
        """
        :returns: The additional number and street address of the address.
        :rtype: str
        """
        return self._properties["street_secondary"]

    def delete(self):
        """
        Deletes the AddressInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._proxy.delete()

    async def delete_async(self):
        """
        Asynchronous coroutine that deletes the AddressInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return await self._proxy.delete_async()

    def fetch(self):
        """
        Fetch the AddressInstance


        :returns: The fetched AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the AddressInstance


        :returns: The fetched AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        return await self._proxy.fetch_async()

    def update(
        self,
        friendly_name=values.unset,
        customer_name=values.unset,
        street=values.unset,
        city=values.unset,
        region=values.unset,
        postal_code=values.unset,
        emergency_enabled=values.unset,
        auto_correct_address=values.unset,
        street_secondary=values.unset,
    ):
        """
        Update the AddressInstance

        :param str friendly_name: A descriptive string that you create to describe the address. It can be up to 64 characters long.
        :param str customer_name: The name to associate with the address.
        :param str street: The number and street address of the address.
        :param str city: The city of the address.
        :param str region: The state or region of the address.
        :param str postal_code: The postal code of the address.
        :param bool emergency_enabled: Whether to enable emergency calling on the address. Can be: `true` or `false`.
        :param bool auto_correct_address: Whether we should automatically correct the address. Can be: `true` or `false` and the default is `true`. If empty or `true`, we will correct the address you provide if necessary. If `false`, we won't alter the address you provide.
        :param str street_secondary: The additional number and street address of the address.

        :returns: The updated AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        return self._proxy.update(
            friendly_name=friendly_name,
            customer_name=customer_name,
            street=street,
            city=city,
            region=region,
            postal_code=postal_code,
            emergency_enabled=emergency_enabled,
            auto_correct_address=auto_correct_address,
            street_secondary=street_secondary,
        )

    async def update_async(
        self,
        friendly_name=values.unset,
        customer_name=values.unset,
        street=values.unset,
        city=values.unset,
        region=values.unset,
        postal_code=values.unset,
        emergency_enabled=values.unset,
        auto_correct_address=values.unset,
        street_secondary=values.unset,
    ):
        """
        Asynchronous coroutine to update the AddressInstance

        :param str friendly_name: A descriptive string that you create to describe the address. It can be up to 64 characters long.
        :param str customer_name: The name to associate with the address.
        :param str street: The number and street address of the address.
        :param str city: The city of the address.
        :param str region: The state or region of the address.
        :param str postal_code: The postal code of the address.
        :param bool emergency_enabled: Whether to enable emergency calling on the address. Can be: `true` or `false`.
        :param bool auto_correct_address: Whether we should automatically correct the address. Can be: `true` or `false` and the default is `true`. If empty or `true`, we will correct the address you provide if necessary. If `false`, we won't alter the address you provide.
        :param str street_secondary: The additional number and street address of the address.

        :returns: The updated AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        return await self._proxy.update_async(
            friendly_name=friendly_name,
            customer_name=customer_name,
            street=street,
            city=city,
            region=region,
            postal_code=postal_code,
            emergency_enabled=emergency_enabled,
            auto_correct_address=auto_correct_address,
            street_secondary=street_secondary,
        )

    @property
    def dependent_phone_numbers(self):
        """
        Access the dependent_phone_numbers

        :returns: twilio.rest.api.v2010.account.address.DependentPhoneNumberList
        :rtype: twilio.rest.api.v2010.account.address.DependentPhoneNumberList
        """
        return self._proxy.dependent_phone_numbers

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.AddressInstance {}>".format(context)


class AddressContext(InstanceContext):
    def __init__(self, version: Version, account_sid: str, sid: str):
        """
        Initialize the AddressContext

        :param Version version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that is responsible for the Address resource to update.
        :param sid: The Twilio-provided string that uniquely identifies the Address resource to update.

        :returns: twilio.rest.api.v2010.account.address.AddressContext
        :rtype: twilio.rest.api.v2010.account.address.AddressContext
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
            "sid": sid,
        }
        self._uri = "/Accounts/{account_sid}/Addresses/{sid}.json".format(
            **self._solution
        )

        self._dependent_phone_numbers: Optional[DependentPhoneNumberList] = None

    def delete(self):
        """
        Deletes the AddressInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._version.delete(
            method="DELETE",
            uri=self._uri,
        )

    async def delete_async(self):
        """
        Asynchronous coroutine that deletes the AddressInstance


        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return await self._version.delete_async(
            method="DELETE",
            uri=self._uri,
        )

    def fetch(self):
        """
        Fetch the AddressInstance


        :returns: The fetched AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """

        payload = self._version.fetch(
            method="GET",
            uri=self._uri,
        )

        return AddressInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the AddressInstance


        :returns: The fetched AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """

        payload = await self._version.fetch_async(
            method="GET",
            uri=self._uri,
        )

        return AddressInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    def update(
        self,
        friendly_name=values.unset,
        customer_name=values.unset,
        street=values.unset,
        city=values.unset,
        region=values.unset,
        postal_code=values.unset,
        emergency_enabled=values.unset,
        auto_correct_address=values.unset,
        street_secondary=values.unset,
    ):
        """
        Update the AddressInstance

        :param str friendly_name: A descriptive string that you create to describe the address. It can be up to 64 characters long.
        :param str customer_name: The name to associate with the address.
        :param str street: The number and street address of the address.
        :param str city: The city of the address.
        :param str region: The state or region of the address.
        :param str postal_code: The postal code of the address.
        :param bool emergency_enabled: Whether to enable emergency calling on the address. Can be: `true` or `false`.
        :param bool auto_correct_address: Whether we should automatically correct the address. Can be: `true` or `false` and the default is `true`. If empty or `true`, we will correct the address you provide if necessary. If `false`, we won't alter the address you provide.
        :param str street_secondary: The additional number and street address of the address.

        :returns: The updated AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        data = values.of(
            {
                "FriendlyName": friendly_name,
                "CustomerName": customer_name,
                "Street": street,
                "City": city,
                "Region": region,
                "PostalCode": postal_code,
                "EmergencyEnabled": emergency_enabled,
                "AutoCorrectAddress": auto_correct_address,
                "StreetSecondary": street_secondary,
            }
        )

        payload = self._version.update(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return AddressInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    async def update_async(
        self,
        friendly_name=values.unset,
        customer_name=values.unset,
        street=values.unset,
        city=values.unset,
        region=values.unset,
        postal_code=values.unset,
        emergency_enabled=values.unset,
        auto_correct_address=values.unset,
        street_secondary=values.unset,
    ):
        """
        Asynchronous coroutine to update the AddressInstance

        :param str friendly_name: A descriptive string that you create to describe the address. It can be up to 64 characters long.
        :param str customer_name: The name to associate with the address.
        :param str street: The number and street address of the address.
        :param str city: The city of the address.
        :param str region: The state or region of the address.
        :param str postal_code: The postal code of the address.
        :param bool emergency_enabled: Whether to enable emergency calling on the address. Can be: `true` or `false`.
        :param bool auto_correct_address: Whether we should automatically correct the address. Can be: `true` or `false` and the default is `true`. If empty or `true`, we will correct the address you provide if necessary. If `false`, we won't alter the address you provide.
        :param str street_secondary: The additional number and street address of the address.

        :returns: The updated AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        data = values.of(
            {
                "FriendlyName": friendly_name,
                "CustomerName": customer_name,
                "Street": street,
                "City": city,
                "Region": region,
                "PostalCode": postal_code,
                "EmergencyEnabled": emergency_enabled,
                "AutoCorrectAddress": auto_correct_address,
                "StreetSecondary": street_secondary,
            }
        )

        payload = await self._version.update_async(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return AddressInstance(
            self._version,
            payload,
            account_sid=self._solution["account_sid"],
            sid=self._solution["sid"],
        )

    @property
    def dependent_phone_numbers(self):
        """
        Access the dependent_phone_numbers

        :returns: twilio.rest.api.v2010.account.address.DependentPhoneNumberList
        :rtype: twilio.rest.api.v2010.account.address.DependentPhoneNumberList
        """
        if self._dependent_phone_numbers is None:
            self._dependent_phone_numbers = DependentPhoneNumberList(
                self._version,
                self._solution["account_sid"],
                self._solution["sid"],
            )
        return self._dependent_phone_numbers

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Api.V2010.AddressContext {}>".format(context)


class AddressPage(Page):
    def get_instance(self, payload):
        """
        Build an instance of AddressInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.api.v2010.account.address.AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        return AddressInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Api.V2010.AddressPage>"


class AddressList(ListResource):
    def __init__(self, version: Version, account_sid: str):
        """
        Initialize the AddressList

        :param Version version: Version that contains the resource
        :param account_sid: The SID of the [Account](https://www.twilio.com/docs/iam/api/account) that is responsible for the Address resource to read.

        :returns: twilio.rest.api.v2010.account.address.AddressList
        :rtype: twilio.rest.api.v2010.account.address.AddressList
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "account_sid": account_sid,
        }
        self._uri = "/Accounts/{account_sid}/Addresses.json".format(**self._solution)

    def create(
        self,
        customer_name,
        street,
        city,
        region,
        postal_code,
        iso_country,
        friendly_name=values.unset,
        emergency_enabled=values.unset,
        auto_correct_address=values.unset,
        street_secondary=values.unset,
    ):
        """
        Create the AddressInstance

        :param str customer_name: The name to associate with the new address.
        :param str street: The number and street address of the new address.
        :param str city: The city of the new address.
        :param str region: The state or region of the new address.
        :param str postal_code: The postal code of the new address.
        :param str iso_country: The ISO country code of the new address.
        :param str friendly_name: A descriptive string that you create to describe the new address. It can be up to 64 characters long.
        :param bool emergency_enabled: Whether to enable emergency calling on the new address. Can be: `true` or `false`.
        :param bool auto_correct_address: Whether we should automatically correct the address. Can be: `true` or `false` and the default is `true`. If empty or `true`, we will correct the address you provide if necessary. If `false`, we won't alter the address you provide.
        :param str street_secondary: The additional number and street address of the address.

        :returns: The created AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        data = values.of(
            {
                "CustomerName": customer_name,
                "Street": street,
                "City": city,
                "Region": region,
                "PostalCode": postal_code,
                "IsoCountry": iso_country,
                "FriendlyName": friendly_name,
                "EmergencyEnabled": emergency_enabled,
                "AutoCorrectAddress": auto_correct_address,
                "StreetSecondary": street_secondary,
            }
        )

        payload = self._version.create(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return AddressInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    async def create_async(
        self,
        customer_name,
        street,
        city,
        region,
        postal_code,
        iso_country,
        friendly_name=values.unset,
        emergency_enabled=values.unset,
        auto_correct_address=values.unset,
        street_secondary=values.unset,
    ):
        """
        Asynchronously create the AddressInstance

        :param str customer_name: The name to associate with the new address.
        :param str street: The number and street address of the new address.
        :param str city: The city of the new address.
        :param str region: The state or region of the new address.
        :param str postal_code: The postal code of the new address.
        :param str iso_country: The ISO country code of the new address.
        :param str friendly_name: A descriptive string that you create to describe the new address. It can be up to 64 characters long.
        :param bool emergency_enabled: Whether to enable emergency calling on the new address. Can be: `true` or `false`.
        :param bool auto_correct_address: Whether we should automatically correct the address. Can be: `true` or `false` and the default is `true`. If empty or `true`, we will correct the address you provide if necessary. If `false`, we won't alter the address you provide.
        :param str street_secondary: The additional number and street address of the address.

        :returns: The created AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressInstance
        """
        data = values.of(
            {
                "CustomerName": customer_name,
                "Street": street,
                "City": city,
                "Region": region,
                "PostalCode": postal_code,
                "IsoCountry": iso_country,
                "FriendlyName": friendly_name,
                "EmergencyEnabled": emergency_enabled,
                "AutoCorrectAddress": auto_correct_address,
                "StreetSecondary": street_secondary,
            }
        )

        payload = await self._version.create_async(
            method="POST",
            uri=self._uri,
            data=data,
        )

        return AddressInstance(
            self._version, payload, account_sid=self._solution["account_sid"]
        )

    def stream(
        self,
        customer_name=values.unset,
        friendly_name=values.unset,
        iso_country=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Streams AddressInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str customer_name: The `customer_name` of the Address resources to read.
        :param str friendly_name: The string that identifies the Address resources to read.
        :param str iso_country: The ISO country code of the Address resources to read.
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.address.AddressInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(
            customer_name=customer_name,
            friendly_name=friendly_name,
            iso_country=iso_country,
            page_size=limits["page_size"],
        )

        return self._version.stream(page, limits["limit"])

    async def stream_async(
        self,
        customer_name=values.unset,
        friendly_name=values.unset,
        iso_country=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Asynchronously streams AddressInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param str customer_name: The `customer_name` of the Address resources to read.
        :param str friendly_name: The string that identifies the Address resources to read.
        :param str iso_country: The ISO country code of the Address resources to read.
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.address.AddressInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(
            customer_name=customer_name,
            friendly_name=friendly_name,
            iso_country=iso_country,
            page_size=limits["page_size"],
        )

        return await self._version.stream_async(page, limits["limit"])

    def list(
        self,
        customer_name=values.unset,
        friendly_name=values.unset,
        iso_country=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Lists AddressInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str customer_name: The `customer_name` of the Address resources to read.
        :param str friendly_name: The string that identifies the Address resources to read.
        :param str iso_country: The ISO country code of the Address resources to read.
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.address.AddressInstance]
        """
        return list(
            self.stream(
                customer_name=customer_name,
                friendly_name=friendly_name,
                iso_country=iso_country,
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(
        self,
        customer_name=values.unset,
        friendly_name=values.unset,
        iso_country=values.unset,
        limit=None,
        page_size=None,
    ):
        """
        Asynchronously lists AddressInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param str customer_name: The `customer_name` of the Address resources to read.
        :param str friendly_name: The string that identifies the Address resources to read.
        :param str iso_country: The ISO country code of the Address resources to read.
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.address.AddressInstance]
        """
        return list(
            await self.stream_async(
                customer_name=customer_name,
                friendly_name=friendly_name,
                iso_country=iso_country,
                limit=limit,
                page_size=page_size,
            )
        )

    def page(
        self,
        customer_name=values.unset,
        friendly_name=values.unset,
        iso_country=values.unset,
        page_token=values.unset,
        page_number=values.unset,
        page_size=values.unset,
    ):
        """
        Retrieve a single page of AddressInstance records from the API.
        Request is executed immediately

        :param str customer_name: The `customer_name` of the Address resources to read.
        :param str friendly_name: The string that identifies the Address resources to read.
        :param str iso_country: The ISO country code of the Address resources to read.
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressPage
        """
        data = values.of(
            {
                "CustomerName": customer_name,
                "FriendlyName": friendly_name,
                "IsoCountry": iso_country,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = self._version.page(method="GET", uri=self._uri, params=data)
        return AddressPage(self._version, response, self._solution)

    async def page_async(
        self,
        customer_name=values.unset,
        friendly_name=values.unset,
        iso_country=values.unset,
        page_token=values.unset,
        page_number=values.unset,
        page_size=values.unset,
    ):
        """
        Asynchronously retrieve a single page of AddressInstance records from the API.
        Request is executed immediately

        :param str customer_name: The `customer_name` of the Address resources to read.
        :param str friendly_name: The string that identifies the Address resources to read.
        :param str iso_country: The ISO country code of the Address resources to read.
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressPage
        """
        data = values.of(
            {
                "CustomerName": customer_name,
                "FriendlyName": friendly_name,
                "IsoCountry": iso_country,
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data
        )
        return AddressPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of AddressInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressPage
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return AddressPage(self._version, response, self._solution)

    async def get_page_async(self, target_url):
        """
        Asynchronously retrieve a specific page of AddressInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of AddressInstance
        :rtype: twilio.rest.api.v2010.account.address.AddressPage
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return AddressPage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a AddressContext

        :param sid: The Twilio-provided string that uniquely identifies the Address resource to update.

        :returns: twilio.rest.api.v2010.account.address.AddressContext
        :rtype: twilio.rest.api.v2010.account.address.AddressContext
        """
        return AddressContext(
            self._version, account_sid=self._solution["account_sid"], sid=sid
        )

    def __call__(self, sid):
        """
        Constructs a AddressContext

        :param sid: The Twilio-provided string that uniquely identifies the Address resource to update.

        :returns: twilio.rest.api.v2010.account.address.AddressContext
        :rtype: twilio.rest.api.v2010.account.address.AddressContext
        """
        return AddressContext(
            self._version, account_sid=self._solution["account_sid"], sid=sid
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return "<Twilio.Api.V2010.AddressList>"



r"""
    This code was generated by
   ___ _ _ _ _ _    _ ____    ____ ____ _    ____ ____ _  _ ____ ____ ____ ___ __   __
    |  | | | | |    | |  | __ |  | |__| | __ | __ |___ |\ | |___ |__/ |__|  | |  | |__/
    |  |_|_| | |___ | |__|    |__| |  | |    |__] |___ | \| |___ |  \ |  |  | |__| |  \

    Twilio - Insights
    This is the public Twilio REST API.

    NOTE: This class is auto generated by OpenAPI Generator.
    https://openapi-generator.tech
    Do not edit the class manually.
"""


from typing import Optional
from twilio.base import deserialize, values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.version import Version
from twilio.base.page import Page


class ParticipantInstance(InstanceResource):
    class Codec(object):
        VP8 = "VP8"
        H264 = "H264"
        VP9 = "VP9"

    class EdgeLocation(object):
        ASHBURN = "ashburn"
        DUBLIN = "dublin"
        FRANKFURT = "frankfurt"
        SINGAPORE = "singapore"
        SYDNEY = "sydney"
        SAO_PAULO = "sao_paulo"
        ROAMING = "roaming"
        UMATILLA = "umatilla"
        TOKYO = "tokyo"

    class RoomStatus(object):
        IN_PROGRESS = "in_progress"
        COMPLETED = "completed"

    class TwilioRealm(object):
        US1 = "us1"
        US2 = "us2"
        AU1 = "au1"
        BR1 = "br1"
        IE1 = "ie1"
        JP1 = "jp1"
        SG1 = "sg1"
        IN1 = "in1"
        DE1 = "de1"
        GLL = "gll"

    def __init__(
        self, version, payload, room_sid: str, participant_sid: Optional[str] = None
    ):
        """
        Initialize the ParticipantInstance

        :returns: twilio.rest.insights.v1.room.participant.ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantInstance
        """
        super().__init__(version)

        self._properties = {
            "participant_sid": payload.get("participant_sid"),
            "participant_identity": payload.get("participant_identity"),
            "join_time": deserialize.iso8601_datetime(payload.get("join_time")),
            "leave_time": deserialize.iso8601_datetime(payload.get("leave_time")),
            "duration_sec": payload.get("duration_sec"),
            "account_sid": payload.get("account_sid"),
            "room_sid": payload.get("room_sid"),
            "status": payload.get("status"),
            "codecs": payload.get("codecs"),
            "end_reason": payload.get("end_reason"),
            "error_code": deserialize.integer(payload.get("error_code")),
            "error_code_url": payload.get("error_code_url"),
            "media_region": payload.get("media_region"),
            "properties": payload.get("properties"),
            "edge_location": payload.get("edge_location"),
            "publisher_info": payload.get("publisher_info"),
            "url": payload.get("url"),
        }

        self._solution = {
            "room_sid": room_sid,
            "participant_sid": participant_sid or self._properties["participant_sid"],
        }
        self._context: Optional[ParticipantContext] = None

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions. All instance actions are proxied to the context

        :returns: ParticipantContext for this ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantContext
        """
        if self._context is None:
            self._context = ParticipantContext(
                self._version,
                room_sid=self._solution["room_sid"],
                participant_sid=self._solution["participant_sid"],
            )
        return self._context

    @property
    def participant_sid(self):
        """
        :returns: Unique identifier for the participant.
        :rtype: str
        """
        return self._properties["participant_sid"]

    @property
    def participant_identity(self):
        """
        :returns: The application-defined string that uniquely identifies the participant within a Room.
        :rtype: str
        """
        return self._properties["participant_identity"]

    @property
    def join_time(self):
        """
        :returns: When the participant joined the room.
        :rtype: datetime
        """
        return self._properties["join_time"]

    @property
    def leave_time(self):
        """
        :returns: When the participant left the room.
        :rtype: datetime
        """
        return self._properties["leave_time"]

    @property
    def duration_sec(self):
        """
        :returns: Amount of time in seconds the participant was in the room.
        :rtype: int
        """
        return self._properties["duration_sec"]

    @property
    def account_sid(self):
        """
        :returns: Account SID associated with the room.
        :rtype: str
        """
        return self._properties["account_sid"]

    @property
    def room_sid(self):
        """
        :returns: Unique identifier for the room.
        :rtype: str
        """
        return self._properties["room_sid"]

    @property
    def status(self):
        """
        :returns:
        :rtype: ParticipantInstance.RoomStatus
        """
        return self._properties["status"]

    @property
    def codecs(self):
        """
        :returns: Codecs detected from the participant. Can be `VP8`, `H264`, or `VP9`.
        :rtype: list[ParticipantInstance.Codec]
        """
        return self._properties["codecs"]

    @property
    def end_reason(self):
        """
        :returns: Reason the participant left the room. See [the list of possible values here](https://www.twilio.com/docs/video/video-log-analyzer/video-log-analyzer-api#end_reason).
        :rtype: str
        """
        return self._properties["end_reason"]

    @property
    def error_code(self):
        """
        :returns: Errors encountered by the participant.
        :rtype: int
        """
        return self._properties["error_code"]

    @property
    def error_code_url(self):
        """
        :returns: Twilio error code dictionary link.
        :rtype: str
        """
        return self._properties["error_code_url"]

    @property
    def media_region(self):
        """
        :returns:
        :rtype: ParticipantInstance.TwilioRealm
        """
        return self._properties["media_region"]

    @property
    def properties(self):
        """
        :returns: Object containing information about the participant's data from the room. See [below](https://www.twilio.com/docs/video/video-log-analyzer/video-log-analyzer-api#properties) for more information.
        :rtype: dict
        """
        return self._properties["properties"]

    @property
    def edge_location(self):
        """
        :returns:
        :rtype: ParticipantInstance.EdgeLocation
        """
        return self._properties["edge_location"]

    @property
    def publisher_info(self):
        """
        :returns: Object containing information about the SDK name and version. See [below](https://www.twilio.com/docs/video/video-log-analyzer/video-log-analyzer-api#publisher_info) for more information.
        :rtype: dict
        """
        return self._properties["publisher_info"]

    @property
    def url(self):
        """
        :returns: URL of the participant resource.
        :rtype: str
        """
        return self._properties["url"]

    def fetch(self):
        """
        Fetch the ParticipantInstance


        :returns: The fetched ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantInstance
        """
        return self._proxy.fetch()

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the ParticipantInstance


        :returns: The fetched ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantInstance
        """
        return await self._proxy.fetch_async()

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Insights.V1.ParticipantInstance {}>".format(context)


class ParticipantContext(InstanceContext):
    def __init__(self, version: Version, room_sid: str, participant_sid: str):
        """
        Initialize the ParticipantContext

        :param Version version: Version that contains the resource
        :param room_sid: The SID of the Room resource.
        :param participant_sid: The SID of the Participant resource.

        :returns: twilio.rest.insights.v1.room.participant.ParticipantContext
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantContext
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "room_sid": room_sid,
            "participant_sid": participant_sid,
        }
        self._uri = "/Video/Rooms/{room_sid}/Participants/{participant_sid}".format(
            **self._solution
        )

    def fetch(self):
        """
        Fetch the ParticipantInstance


        :returns: The fetched ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantInstance
        """

        payload = self._version.fetch(
            method="GET",
            uri=self._uri,
        )

        return ParticipantInstance(
            self._version,
            payload,
            room_sid=self._solution["room_sid"],
            participant_sid=self._solution["participant_sid"],
        )

    async def fetch_async(self):
        """
        Asynchronous coroutine to fetch the ParticipantInstance


        :returns: The fetched ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantInstance
        """

        payload = await self._version.fetch_async(
            method="GET",
            uri=self._uri,
        )

        return ParticipantInstance(
            self._version,
            payload,
            room_sid=self._solution["room_sid"],
            participant_sid=self._solution["participant_sid"],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = " ".join("{}={}".format(k, v) for k, v in self._solution.items())
        return "<Twilio.Insights.V1.ParticipantContext {}>".format(context)


class ParticipantPage(Page):
    def get_instance(self, payload):
        """
        Build an instance of ParticipantInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.insights.v1.room.participant.ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantInstance
        """
        return ParticipantInstance(
            self._version, payload, room_sid=self._solution["room_sid"]
        )

    def __repr__(self) -> str:
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        """
        return "<Twilio.Insights.V1.ParticipantPage>"


class ParticipantList(ListResource):
    def __init__(self, version: Version, room_sid: str):
        """
        Initialize the ParticipantList

        :param Version version: Version that contains the resource
        :param room_sid: The SID of the Room resource.

        :returns: twilio.rest.insights.v1.room.participant.ParticipantList
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantList
        """
        super().__init__(version)

        # Path Solution
        self._solution = {
            "room_sid": room_sid,
        }
        self._uri = "/Video/Rooms/{room_sid}/Participants".format(**self._solution)

    def stream(self, limit=None, page_size=None):
        """
        Streams ParticipantInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.insights.v1.room.participant.ParticipantInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = self.page(page_size=limits["page_size"])

        return self._version.stream(page, limits["limit"])

    async def stream_async(self, limit=None, page_size=None):
        """
        Asynchronously streams ParticipantInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.insights.v1.room.participant.ParticipantInstance]
        """
        limits = self._version.read_limits(limit, page_size)
        page = await self.page_async(page_size=limits["page_size"])

        return await self._version.stream_async(page, limits["limit"])

    def list(self, limit=None, page_size=None):
        """
        Lists ParticipantInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.insights.v1.room.participant.ParticipantInstance]
        """
        return list(
            self.stream(
                limit=limit,
                page_size=page_size,
            )
        )

    async def list_async(self, limit=None, page_size=None):
        """
        Asynchronously lists ParticipantInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.insights.v1.room.participant.ParticipantInstance]
        """
        return list(
            await self.stream_async(
                limit=limit,
                page_size=page_size,
            )
        )

    def page(
        self, page_token=values.unset, page_number=values.unset, page_size=values.unset
    ):
        """
        Retrieve a single page of ParticipantInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantPage
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = self._version.page(method="GET", uri=self._uri, params=data)
        return ParticipantPage(self._version, response, self._solution)

    async def page_async(
        self, page_token=values.unset, page_number=values.unset, page_size=values.unset
    ):
        """
        Asynchronously retrieve a single page of ParticipantInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantPage
        """
        data = values.of(
            {
                "PageToken": page_token,
                "Page": page_number,
                "PageSize": page_size,
            }
        )

        response = await self._version.page_async(
            method="GET", uri=self._uri, params=data
        )
        return ParticipantPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of ParticipantInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantPage
        """
        response = self._version.domain.twilio.request("GET", target_url)
        return ParticipantPage(self._version, response, self._solution)

    async def get_page_async(self, target_url):
        """
        Asynchronously retrieve a specific page of ParticipantInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of ParticipantInstance
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantPage
        """
        response = await self._version.domain.twilio.request_async("GET", target_url)
        return ParticipantPage(self._version, response, self._solution)

    def get(self, participant_sid):
        """
        Constructs a ParticipantContext

        :param participant_sid: The SID of the Participant resource.

        :returns: twilio.rest.insights.v1.room.participant.ParticipantContext
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantContext
        """
        return ParticipantContext(
            self._version,
            room_sid=self._solution["room_sid"],
            participant_sid=participant_sid,
        )

    def __call__(self, participant_sid):
        """
        Constructs a ParticipantContext

        :param participant_sid: The SID of the Participant resource.

        :returns: twilio.rest.insights.v1.room.participant.ParticipantContext
        :rtype: twilio.rest.insights.v1.room.participant.ParticipantContext
        """
        return ParticipantContext(
            self._version,
            room_sid=self._solution["room_sid"],
            participant_sid=participant_sid,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return "<Twilio.Insights.V1.ParticipantList>"
